/*
 * HP Model.
 *
 *   Shamelessly based on Jean Delvare's DMI Decode. This source is
 *  distributed freely and should work on x86, x86_64 and ia64 platforms
 *  returning information about the server on which the Linux operating
 *  system is running. Since DMI decode is GPL 2 this 
 *  source is also released under the GNU Public License version 2.
 *  For the original and much more complete code, obtain the dmidecode from
 *  http://www.nongnu.org/dmidecode/. Enjoy. 
 *       ----------------------------------------------
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 *   For the avoidance of doubt the "preferred form" of this code is one which
 *   is in an open unpatent encumbered format. Where cryptographic key signing
 *   forms part of the process of creating an executable the information 
 *   including keys needed to generate an equivalently functional executable
 *   are deemed to be part of the source code.
 *
 * Unless specified otherwise, all references are aimed at the "System
 * Management BIOS Reference Specification, Version 2.3.4" document,
 * available from http://www.dmtf.org/standards/bios.php.
 *
 * Note to contributors:
 * Please reference every value you add or modify, especially if the
 * information does not come from the above mentioned specification.
 *
 * Additional references:
 *	- Intel AP-485 revision 23
 *    "Intel Processor Identification and the CPUID Instruction"
 *    http://developer.intel.com/design/xeon/applnots/241618.htm
 *  - DMTF Master MIF version 021205
 *    "DMTF approved standard groups"
 *    http://www.dmtf.org/standards/standard_dmi.php
 *  - IPMI 1.5 revision 1.1
 *    "Intelligent Platform Management Interface Specification"
 *    http://developer.intel.com/design/servers/ipmi/spec.htm
 */

#include <sys/utsname.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <byteswap.h>

#ifdef USE_MMAP
#include <sys/mman.h>
#endif /* USE MMAP */
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>

#ifdef __ia64
#include <byteswap.h>
#define ALIGNMENT_WORKAROUND
#endif

#include "version.h"
#include "types.h"
#include "util.h"

#define HPMODEL_VERSION	FQVERSION

#define VIRTUAL_ID_DISABLED    "Virtual IDs not enabled"
#define VIRTUAL_ID_UNSUPPORTED "No Virtual ID support"

/* For command line option print modes */
#define PRINT_PLATFORM_TYPE	0x1
#define PRINT_MFG		0x2
#define PRINT_PRODNAME		0x4
#define PRINT_PRODREV		0x8
#define PRINT_SERIAL		0x10
#define PRINT_UUID		0x20
#define PRINT_WAKEUP_MODE	0x40
#define PRINT_VIRTUALS	        0x80
#define PRINT_SMBIOS_VERSION	0x100
#define PRINT_ALL		0xFFFF

/* global print flag for what info the user needs. */
int print_mode=0;  

static const char *out_of_spec = "<OUT OF SPEC>";
static const char *bad_index = "<BAD INDEX>";

/*
 * The specification isn't very clear on endianness problems, so we better
 * have macros for these. It also helps us solve problems on systems that
 * don't support non-aligned memory access. This isn't a big issue IMHO,
 * since SMBIOS/DMI is intended mainly for Intel and compatible systems,
 * which are little-endian and support non-aligned memory access. Anyway,
 * you may use the following defines to control the way it works:
 * - Define BIGENDIAN on big-endian systems.
 * - Define ALIGNMENT_WORKAROUND if your system doesn't support
 *   non-aligned memory access. In this case, we use a slower, but safer,
 *   memory access method.
 * - If it happens that the table is supposed to be always little-endian
 *   ordered regardless of the architecture, define TABLE_LITTLEENDIAN.
 * You most probably will have to define none or the three of them.
 */

#ifndef TABLE_LITTLEENDIAN
#	ifdef BIGENDIAN
	typedef struct {
		u32 h;
		u32 l;
	} u64;
#	else /* BIGENDIAN */
	typedef struct {
		u32 l;
		u32 h;
	} u64;
#	endif /* BIGENDIAN */
#	ifdef ALIGNMENT_WORKAROUND
#		ifdef BIGENDIAN
#		define WORD(x) (u16)((x)[1]+((x)[0]<<8))
#		define DWORD(x) (u32)((x)[3]+((x)[2]<<8)+((x)[1]<<16)+((x)[0]<<24))
#		define QWORD(x) (U64(DWORD(x+4), DWORD(x)))
#		else /* BIGENDIAN */
#		define WORD(x) (u16)((x)[0]+((x)[1]<<8))
#		define DWORD(x) (u32)((x)[0]+((x)[1]<<8)+((x)[2]<<16)+((x)[3]<<24))
#		define QWORD(x) (U64(DWORD(x), DWORD(x+4)))
#		endif /* BIGENDIAN */
#	else /* ALIGNMENT_WORKAROUND */
#	define WORD(x) (u16)(*(u16 *)(x))
#	define DWORD(x) (u32)(*(u32 *)(x))
#	define QWORD(x) (*(u64 *)(x))
#	endif /* ALIGNMENT_WORKAROUND */
#else /* TABLE_LITTLEENDIAN */
typedef struct {
	u32 l;
	u32 h;
} u64;
#define WORD(x) (u16)((x)[0]+((x)[1]<<8))
#define DWORD(x) (u32)((x)[0]+((x)[1]<<8)+((x)[2]<<16)+((x)[3]<<24))
#define QWORD(x) (U64(DWORD(x), DWORD(x+4)))
#endif /* TABLE_LITTLEENDIAN */

struct dmi_header
{
	u8 type;
	u8 length;
	u16 handle;
};

struct system_info {
        char mfg[32];
        char productName[32];
        char productVer[32];
        char platform[32];
        char serial[32];
        char uuid[37];
        char wakeupType[32];
        char v_uuid[37];
        char v_serial[32];
};

static struct system_info sysInfo;

#if ((defined BIGENDIAN && defined TABLE_LITTLEENDIAN) || defined ALIGNMENT_WORKAROUND)
#define HANDLE(x) WORD((u8 *)&(x->handle))
#else
#define HANDLE(x) x->handle
#endif

/* Simple function to uppercase the UUID and Serial Number */
static void ToUpper(char *buf)
{
  while (*buf) {
    if (islower(*buf))
      *buf = toupper(*buf);
    buf++;
  }
}

/*
 * Type-independant Stuff
 */

static const char *dmi_string(struct dmi_header *dm, u8 s)
{
	char *bp=(char *)dm;
	size_t i;

	if(s==0)
		return "Not Specified";
	
	bp+=dm->length;
	while(s>1 && *bp)
	{
		bp+=strlen(bp);
		bp++;
		s--;
	}
	
	if(!*bp)
		return bad_index;
	
	/* ASCII filtering */
	for(i=0; i<strlen(bp); i++)
		if(bp[i]<32 || bp[i]==127)
			bp[i]='.';
	
	return bp;
}

/*
 * 3.3.2 System Information (Type 1)
 */

/* Some early HP machines violated the RFC4122 MSB/Network Byte Order
   convention, and store directly in little-endian order. */

/* Excerpted from RFC4122

** Copyright (c) 1989 by Hewlett-Packard Company, Palo Alto, Ca. &
**                       Digital Equipment Corporation, Maynard, Mass.
** Copyright (c) 1990- 1993, 1996 Open Software Foundation, Inc.
** Copyright (c) 1998 Microsoft.

*/

typedef struct {
    u32  time_low;
    u16  time_mid;
    u16  time_hi_and_version;
    u8   clock_seq_hi_and_reserved;
    u8   clock_seq_low;
    u8   node[6];
} __attribute__((packed)) RFC4122uuid_t;



static char * dmi_system_uuid(u8 *p)
{
	RFC4122uuid_t uuid;
	static char uuid_str[37];
	int only0xFF=1, only0x00=1;
	int i;
	u16 magic;

	/* gdb can't hit *p directly because it's memory-mapped file space */

	memcpy(uuid_str, p, 16);	/* gdb: p /x *uuid_str@16 */
	memcpy(&uuid, p, 16);

	/* Check the magic field.  If it works, it's the old, technically
	   broken storage order, but I really want a string representation,
	   so just use it as it. */

	magic = uuid.time_hi_and_version;
	if (magic && ((magic & 0x11D0) != 0x11D0)) {
	    if ((magic & 0xD011) == 0xD011) {	/* stable storage is RFC4122 */
    		uuid.time_low 		 = bswap_32(uuid.time_low);
		uuid.time_mid 		 = bswap_16(uuid.time_mid);
		uuid.time_hi_and_version = bswap_16(uuid.time_hi_and_version);
	    } else {
               #ifdef DEBUG
	    	  printf("Virtual UUID Found\n");
               #endif
	    }
	}

	for(i=0; i<16 && (only0x00 || only0xFF); i++)
	{
		if(p[i]!=0x00) only0x00=0;
		if(p[i]!=0xFF) only0xFF=0;
	}
	
	if(only0xFF)
	{
		sprintf(uuid_str, "Not Present");
		return uuid_str;
	}
	if(only0x00)
	{
		sprintf(uuid_str, "Not Settable");
		return uuid_str;
	}

	sprintf(uuid_str, "%08x-%04x-%04x-"
			  "%02x%02x-"
			  "%02x%02x%02x%02x%02x%02x",
	    uuid.time_low, uuid.time_mid, uuid.time_hi_and_version, 
	    uuid.clock_seq_hi_and_reserved, uuid.clock_seq_low, 
	    uuid.node[0], uuid.node[1], uuid.node[2], 
	    uuid.node[3], uuid.node[4], uuid.node[5]);
        ToUpper(uuid_str);
	return uuid_str;
}

static const char *dmi_system_wake_up_type(u8 code)
{
	/* 3.3.2.1 */
	static const char *type[]={
		"Reserved", /* 0x00 */
		"Other",
		"Unknown",
		"APM Timer",
		"Modem Ring",
		"LAN Remote",
		"Power Switch",
		"PCI PME#",
		"AC Power Restored" /* 0x08 */
	};
	
	if(code<=0x08)
		return type[code];
	return out_of_spec;
}

/*
 * Main
 */

/*
 *  strchop - chops trailing white space.
 */
char *strchop (const char *str_to_chop)
{

   char *c,*ret;
   ret = strdup(str_to_chop);
   c = &ret[strlen(ret)-1];
   while (c != ret && isspace(*c)) {
	*c = '\0';
	c--;
   }
   return ret;
}

static char * get_arch(void)
{
   static char platform[10];
   struct utsname utsbuf;

   if (strlen(platform) < 1) {
	if (uname( &utsbuf )) {
		sprintf(platform, "unknown");
	} else {
		sprintf(platform, "%s", utsbuf.machine );
	}
   }
   return platform;
}

static char SMBIOSversion[8] = "";	/* Simplest solution */

/*
 * print_info: Based on global print_mode flags, prints out various pieces of 
 *          sysInfo.
 * returns:  None
 */
void print_info(void) {
   if (print_mode == PRINT_ALL) {
   printf("%s:\n\n\
Manufacturer Name:      %s\n\
Product Name:           %s\n\
Product Version:        %s\n\
Hardware Platform:      %s\n\
Physical Serial Number: %s\n\
Physical UUID:          %s\n\
System Wake-up Type:    %s\n\
Virtual Serial Number:  %s\n\
Virtual UUID:           %s\n\
SMBIOS Version:         %s\n",
     HPMODEL_VERSION,
   sysInfo.mfg,
   sysInfo.productName,
   sysInfo.productVer,
   sysInfo.platform,
   sysInfo.serial,
   sysInfo.uuid,
   sysInfo.wakeupType,
   sysInfo.v_serial,
   sysInfo.v_uuid,
   SMBIOSversion);

   } else { 
      if (print_mode & PRINT_PLATFORM_TYPE)
         printf ("%s ", sysInfo.platform);
      if (print_mode & PRINT_MFG)
         printf ("%s ", sysInfo.mfg);
      if (print_mode & PRINT_PRODNAME)
         printf ("%s ", sysInfo.productName);
      if (print_mode & PRINT_PRODREV)
         printf ("%s ", sysInfo.productVer);
      if (print_mode & PRINT_SERIAL)
         printf ("%s ", sysInfo.serial);
      if (print_mode & PRINT_UUID)
         printf ("%s ", sysInfo.uuid);
      if (print_mode & PRINT_VIRTUALS) 
         printf ("%s %s", sysInfo.v_uuid, sysInfo.v_serial);
      if (print_mode & PRINT_WAKEUP_MODE)
         printf ("%s ", sysInfo.wakeupType);
      if (print_mode & PRINT_SMBIOS_VERSION)
         printf ("%s ", SMBIOSversion);
      printf ("\n");
   }
}


/*
 * dmi_decode: decodes dmi_header and data for the dmi tables. Specifically,
 *          decodes the system table (type 1). If we wanted to support mining
 *          DMI data from other table types we will add additional cases in 
 *          the future.
 * returns:  1 upon successful decode of the DMI table type 1. 
 *           returns 0 otherwise
 */
static int dmi_decode(u8 *data)
{
        char buf_serial[32];
	struct dmi_header *h=(struct dmi_header *)data;
	
	switch(h->type)
	{
           case 226:   /* This OEM table type signifies HP's Physical Serial# 
                        * and physical UUID. If system virtualUUIDs are 
                        * enabled, the Physical UUID/SerialNumber will
                        * be different than the virtualUUID/SerialNum found
                        * in SMBIOS table type 1 */
              strcpy(buf_serial, strchop(dmi_string(h,data[0x14])));
              if(strcmp(buf_serial, sysInfo.serial) == 0) {
                 /* then we are looking at the real uuid/serial NOW */
                 strcpy(sysInfo.uuid, dmi_system_uuid(data+0x4));
                 strcpy(sysInfo.v_serial, VIRTUAL_ID_DISABLED);
                 strcpy(sysInfo.v_uuid, VIRTUAL_ID_DISABLED);
              } else {
                 strcpy(sysInfo.v_uuid, sysInfo.uuid);
                 strcpy(sysInfo.v_serial, sysInfo.serial);
                 ToUpper(sysInfo.v_serial);   /* Just in CASE */
                 strcpy(sysInfo.serial, buf_serial);
                 ToUpper(sysInfo.serial);     /* Just in CASE */
                 strcpy(sysInfo.uuid, dmi_system_uuid(data+0x4));
              }
              break;
	   case 1: /* 3.3.2 System Information */
              strcpy(sysInfo.mfg, strchop(dmi_string(h, data[0x04])));
              strcpy(sysInfo.productName, strchop(dmi_string(h, data[0x05])));
              strcpy(sysInfo.productVer, strchop(dmi_string(h, data[0x06])));
              strcpy(sysInfo.platform, get_arch());
              strcpy(sysInfo.serial, strchop(dmi_string(h, data[0x07])));
              strcpy(sysInfo.uuid, dmi_system_uuid(data+0x8));
              strcpy(sysInfo.wakeupType, dmi_system_wake_up_type(data[0x18]));
              strcpy(sysInfo.v_uuid, VIRTUAL_ID_UNSUPPORTED);
              strcpy(sysInfo.v_serial, VIRTUAL_ID_UNSUPPORTED);
	      break;
	   default:
	      break;
        }
	return 0;
}
		
#ifdef USE_MMAP
static void dmi_table(int fd, u32 base, u16 len, u16 num, __attribute__ ((unused)) const char *pname, __attribute__ ((unused)) const char *devmem)
#else /* USE_MMAP */
static void dmi_table(int fd, u32 base, u16 len, u16 num, const char *pname, const char *devmem)
#endif /* USE_MMAP */
{
	u8 *buf;
	u8 *data;
	int i=0;
#ifdef USE_MMAP
	u32    mmquot;
	u32    mmrem;
	size_t psize;
	void *mmp;
#endif /* USE_MMAP */
	
#ifdef USE_MMAP
	psize=getpagesize();
	mmquot=base / psize;
	mmrem =base % psize;
	/*
	 * We need PROT_WRITE for ASCII filtering in dmi_string. Do NOT change MAP_PRIVATE
	 * to MAP_SHARED unless you also remove PROT_WRITE and disable ASCII filtering, because
	 * we DO NOT WANT to write to /dev/mem. Thanks a lot to Gabriel Forte for helping
	 * me hunting down an annoying bug related to this.
	 */
	mmp=mmap(0, mmrem+len, PROT_READ|PROT_WRITE, MAP_PRIVATE, fd, mmquot*psize);
    if(mmp==MAP_FAILED)
    {
       perror("mmap");
       return;
    }
	buf=(u8 *)mmp+mmrem;
#else /* USE_MMAP */
	if((buf=malloc(len))==NULL)
	{
		perror(pname);
		return;
	}
	if(lseek(fd, (off_t)base, SEEK_SET)==-1)
	{
		perror(devmem);
		return;
	}
	if(myread(fd, buf, len, devmem)==-1)
	{
		free(buf);
		printf("Table is unreachable, sorry. Try compiling dmidecode with -DUSE_MMAP.\n"
			"This problem is known on the IBM T23, T30 and X30 laptops, the Fujitsu-Siemens\n"
			"S-4582 laptop as well as IA-64 systems. If your system differ, please report\n");
		exit(1);
	}
#endif /* USE_MMAP */
	
	data=buf;
	while(i<num && data+sizeof(struct dmi_header)<=buf+len)
	{
		u8 *next;
		struct dmi_header *h=(struct dmi_header *)data;
		
		/* look for the next handle */
		next=data+h->length;
		while(next-buf+1<len && (next[0]!=0 || next[1]!=0))
			next++;
		next+=2;
		if(next-buf<=len)
		   dmi_decode(data);
		
		data=next;
		i++;
	} 

        if (print_mode == PRINT_VIRTUALS &&
            ((strcmp(sysInfo.v_uuid, VIRTUAL_ID_UNSUPPORTED)==0) || 
             (strcmp(sysInfo.v_uuid, VIRTUAL_ID_DISABLED)==0))) 
              printf("%s", sysInfo.v_uuid);
        else 
           print_info();
          
        
	
#ifdef USE_MMAP
	if(munmap(mmp, mmrem+len)==-1)
		perror("munmap");
#else /* USE_MMAP */
	free(buf);
#endif /* USE_MMAP */
}

static int smbios_decode(u8 *buf, int fd, const char *pname, const char *devmem)
{
	if(checksum(buf, buf[0x05])
	 && memcmp(buf+0x10, "_DMI_", 5)==0
	 && checksum(buf+0x10, 0x0F))
	{
                sprintf(SMBIOSversion, "%u.%u", buf[0x06], buf[0x07]);

		dmi_table(fd, DWORD(buf+0x18), WORD(buf+0x16), WORD(buf+0x1C),
			pname, devmem);
		return 1;
	}
	
	return 0;
}

void usage(const char *prog) {
   printf("Usage: %s [OPTION]...\n", prog);
   printf("\
Print hardware information about this system obtained from SMBIOS.\n\n\
With no OPTION, same as -imp.\n\
	-a           print all information from SMBIOS system table\n\
	-i 	     print system platform type\n\
	-m	     print system manufacturer\n\
	-p	     print system product name\n\
	-r	     print system version\n\
	-s	     print system serial number\n\
	-u	     print system UUID\n\
	-w	     print system Wake up type\n\
	-v	     print SMBIOS version\n\n");
	exit(1);
}


int main(int argc, char *argv[])
{
	int fd,i;
	off_t fp=0xF0000;
	const char *devmem="/dev/mem";
#ifdef __ia64
	FILE *efi_systab;
	const char *filename;
	char linebuf[64];
#ifdef USE_MMAP
	div_t mmbase;
	size_t psize;
	void *mmp;
#else /* USE_MMAP */
	u8 buf[0x20];
#endif /* USE_MMAP */
#else /* __ia64 */
	u8 buf[0x20];
#endif /* __ia64 */
	
	if(sizeof(u8)!=1 || sizeof(u16)!=2 || sizeof(u32)!=4 || '\0'!=0)
	{
		fprintf(stderr,"%s: compiler incompatibility\n", argv[0]);
		exit(255);
	}
	
        while ( ( i = getopt(argc, argv, "aimprstuwv?")) != EOF ) switch(i) {
	   case 'a': print_mode = PRINT_ALL; break;
	   case 'i': print_mode |= PRINT_PLATFORM_TYPE; break;
	   case 'm': print_mode |= PRINT_MFG; break;
	   case 'p': print_mode |= PRINT_PRODNAME; break;
	   case 'r': print_mode |= PRINT_PRODREV; break;
	   case 's': print_mode |= PRINT_SERIAL; break;
	   case 't': print_mode |= PRINT_VIRTUALS; break;
	   case 'u': print_mode |= PRINT_UUID; break;
	   case 'w': print_mode |= PRINT_WAKEUP_MODE; break;
	   case 'v': print_mode |= PRINT_SMBIOS_VERSION; break;
	   case '?': usage(argv[0]);
	}
	if (print_mode == 0) print_mode = PRINT_PLATFORM_TYPE | PRINT_MFG | PRINT_PRODNAME;

	if((fd=open(devmem, O_RDONLY))==-1)
	{
		perror(devmem);
		exit(1);
	}
	
#ifdef __ia64
	if((efi_systab=fopen(filename="/proc/efi/systab", "r"))==NULL
	&& (efi_systab=fopen(filename="/sys/firmware/efi/systab", "r"))==NULL)
	{
		perror(filename);
		exit(1);
	}
	while((fgets(linebuf, sizeof(linebuf)-1, efi_systab))!=NULL)
	{
		char* addr=memchr(linebuf, '=', strlen(linebuf));
		*(addr++)='\0';
		if(strcmp(linebuf, "SMBIOS")==0)
		{
			fp=strtol(addr, NULL, 0);
			#ifdef DEBUG
			printf("# SMBIOS entry point at 0x%08lx\n", fp);
			#endif
		}
	}
	if(fclose(efi_systab)!=0)
		perror(filename);

#ifdef USE_MMAP
	psize=getpagesize();
	mmbase=div(fp, psize);
	mmp=mmap(0, mmbase.rem+0x20, PROT_READ, MAP_PRIVATE, fd, mmbase.quot*psize);
    if(mmp==MAP_FAILED)
    {
       perror(devmem);
       exit(1);
    }

	smbios_decode(((u8 *)mmp)+mmbase.rem, fd, argv[0], devmem);

	if(munmap(mmp, mmbase.rem+0x20)==-1)
		perror("munmap");
#else /* USE_MMAP */
	if(lseek(fd, fp, SEEK_SET)==-1)
	{
		perror(devmem);
		exit(1);
	}
	if(myread(fd, buf, 0x20, devmem)==-1)
		exit(1);

	smbios_decode(buf, fd, argv[0], devmem);
#endif /* USE_MMAP */
#else /* __ia64 */
	if(lseek(fd, fp, SEEK_SET)==-1)
	{
		perror(devmem);
		exit(1);
	}
	while(fp<=0xFFFF0)
	{
		if(myread(fd, buf, 0x10, devmem)==-1)
			exit(1);
		fp+=16;
		
		if(memcmp(buf, "_SM_", 4)==0 && fp<=0xFFFF0)
		{
			if(myread(fd, buf+0x10, 0x10, devmem)==-1)
				exit(1);
			fp+=16;
			
			if(smbios_decode(buf, fd, argv[0], devmem))
			{
				/* dmi_table moved us far away */
				lseek(fd, fp, SEEK_SET);
			}
		}
		else if(memcmp(buf, "_DMI_", 5)==0
		 && checksum(buf, 0x0F))
		{
			printf("Legacy DMI %u.%u present.\n",
				buf[0x0E]>>4, buf[0x0E]&0x0F);
			dmi_table(fd, DWORD(buf+0x08), WORD(buf+0x06), WORD(buf+0x0C),
				argv[0], devmem);
			
			/* dmi_table moved us far away */
			lseek(fd, fp, SEEK_SET);
		}
	}
#endif /* __ia64 */
	
	if(close(fd)==-1)
	{
		perror(devmem);
		exit(1);
	}
	
	return 0;
}
